"""Utilities."""

from collections import defaultdict
from pathlib import Path
import re
import tomli


DEFAULT_CONFIG = {
    "duplicates": set(),
    "skips": set(),
}
FIGURE_DEF = re.compile(r'<figure\s+id="(.+?)"\s*>\s+<img\s+src="(.+?)"\s+alt="(.+?)"\s*>\s+<figcaption>(.+?)</figcaption>\s+</figure>', re.MULTILINE)
KEY_DEF = re.compile(r'^<span\s+id="(.+?)">.+?</span>\s*$', re.MULTILINE)
MD_LINK_DEF = re.compile(r"^\[(.+?)\]:\s+(.+?)\s*$", re.MULTILINE)
SUFFIXES_BIN = {".ico", ".jpg", ".png"}
SUFFIXES_SRC = {".css", ".html", ".js", ".md", ".py", ".sh", ".sql", ".txt"}
SUFFIXES_TXT = SUFFIXES_SRC | {".csv", ".json", ".svg"}
SUFFIXES = SUFFIXES_BIN | SUFFIXES_TXT
TABLE_DEF = re.compile(r'<table\s+id="(.+?)"\s*>.+?<caption>(.+?)</caption>\s*</table>', re.MULTILINE + re.DOTALL)


def find_figure_defs(files):
    """Collect all figure definitions."""
    found = defaultdict(list)
    for filepath, content in files.items():
        if filepath.suffix == ".md":
            for figure in FIGURE_DEF.finditer(content):
                found[figure[1]].append({
                    "img": figure[2],
                    "alt": figure[3],
                    "caption": figure[4],
                })
    return found


def find_files(opt, skips=None):
    """Collect all interesting files."""
    return {
        filepath: read_file(filepath)
        for filepath in Path(opt.root).glob("**/*.*")
        if _is_interesting_file(filepath, skips)
    }


def find_key_defs(files, term):
    """Find key definitions in definition list file."""
    candidates = [k for k in files if term in str(k).lower()]
    if len(candidates) != 1:
        return None
    file_key = candidates[0]
    return set(KEY_DEF.findall(files[file_key]))


def find_table_defs(files):
    """Collect all table definitions."""
    found = defaultdict(list)
    for filepath, content in files.items():
        if filepath.suffix == ".md":
            for table in TABLE_DEF.finditer(content):
                found[table[1]].append({"caption": table[2],})
    return found


def load_config(config_path):
    """Load configuration file or construct default."""
    if config_path is None:
        return DEFAULT_CONFIG
    config = tomli.loads(Path(config_path).read_text())
    if ("tool" not in config) or ("mccole" not in config["tool"]):
        print(f"configuration file {config_path} does not have 'tool.mccole' section")
        return DEFAULT_CONFIG
    config = config["tool"]["mccole"]
    if "duplicates" in config:
        config["duplicates"] = set(frozenset(v) for v in config["duplicates"])
    if "skips" in config:
        config["skips"] = set(config["skips"])
    return config


def read_file(filepath):
    """Read file as bytes or text."""
    if filepath.suffix in SUFFIXES_TXT:
        return filepath.read_text()
    else:
        return filepath.read_bytes()


def write_file(filepath, content):
    """Write file as bytes or text."""
    filepath.parent.mkdir(parents=True, exist_ok=True)
    if filepath.suffix in SUFFIXES_TXT:
        return filepath.write_text(content)
    else:
        return filepath.write_bytes(content)


def _is_interesting_file(filepath, skips):
    """Is this file worth checking?"""
    if not filepath.is_file():
        return False
    if str(filepath).startswith("."):
        return False
    if filepath.suffix not in SUFFIXES:
        return False
    if str(filepath.parent.name).startswith("."):
        return False
    if skips and any(str(filepath).startswith(s) for s in skips):
        return False
    return True



"""Report site stats."""

import argparse
from prettytable import PrettyTable
import re

from .util import FIGURE_DEF, MD_LINK_DEF, find_figure_defs, find_files, find_key_defs, find_table_defs


TABLE_FMT = {
    "stat": "l",
    "value": "r",
}


def stats(opt):
    """Main driver."""
    table = PrettyTable(field_names=TABLE_FMT.keys())
    for k, v in TABLE_FMT.items():
        table.align[k] = v

    files = find_files(opt, set(["bin", opt.out]))
    table.add_row(("bibliography entries", len(find_key_defs(files, "bibliography"))))
    table.add_row(("glossary entries", len(find_key_defs(files, "glossary"))))
    table.add_row(("figures", len(find_figure_defs(files))))
    table.add_row(("link definitions", len(find_markdown_link_defs(files))))
    table.add_row(("tables", len(find_table_defs(files))))

    print(table)


def find_markdown_link_defs(files):
    """Collect Markdown link key definnitions."""
    found = set()
    for filepath, content in files.items():
        if filepath.suffix == ".md":
            for link in MD_LINK_DEF.finditer(content):
                found.add(link[0])
    return found


def parse_args(parser):
    """Parse command-line arguments."""
    parser.add_argument("--out", type=str, default="docs", help="output directory")
    parser.add_argument("--root", type=str, default=".", help="root directory")
    return parser


if __name__ == "__main__":
    opt = parse_args(argparse.ArgumentParser()).parse_args()
    stats(opt)



"""Convert Markdown to HTML."""

import argparse
from bs4 import BeautifulSoup
from jinja2 import Environment, FileSystemLoader
from markdown import markdown
from pathlib import Path

from .util import find_files, load_config, write_file


MARKDOWN_EXTENSIONS = ["attr_list", "def_list", "fenced_code", "md_in_html", "tables"]


def render(opt):
    """Main driver."""
    config = load_config(opt.config)
    skips = config["skips"] | {opt.out}
    files = find_files(opt, skips)
    files = {key:{"content": val} for key, val in files.items()}
    env = Environment(loader=FileSystemLoader(opt.templates))

    for filepath, info in files.items():
        if filepath.suffix == ".md":
            info["doc"] = render_markdown(env, opt, filepath, info["content"])

    for filepath, info in files.items():
        result = str(info["doc"]) if filepath.suffix == ".md" else info["content"]
        output_path = make_output_path(opt.out, config["renames"], filepath)
        write_file(output_path, result)


def choose_template(env, source_path):
    """Select a template."""
    if source_path.name == "slides.md":
        return env.get_template("slides.html")
    return env.get_template("page.html")


def do_bibliography_links(doc, source_path):
    """Turn 'b:key' links into bibliography references."""
    for node in doc.select("a[href]"):
        if node["href"].startswith("b:"):
            node["href"] = f"@root/bibliography.html#{node['href'][2:]}"


def do_glossary_links(doc, source_path):
    """Turn 'g:key' links into glossary references."""
    for node in doc.select("a[href]"):
        if node["href"].startswith("g:"):
            node["href"] = f"@root/glossary.html#{node['href'][2:]}"


def do_markdown_links(doc, source_path):
    """Fix .md links in HTML."""
    for node in doc.select("a[href]"):
        if node["href"].endswith(".md"):
            node["href"] = node["href"].replace(".md", ".html").lower()


def do_tables(doc, source_path):
    """Eliminate duplicate table tags created by Markdown tables inside HTML tables."""
    for node in doc.select("table"):
        parent = node.parent
        if parent.name == "table":
            caption = parent.caption
            node.append(caption)
            parent.replace_with(node)


def do_title(doc, source_path):
    """Make sure title element is filled in."""
    doc.title.string = doc.h1.get_text()


def do_root_path_prefix(doc, source_path):
    """Fix @root links in HTML."""
    depth = len(source_path.parents) - 1
    prefix = "./" if (depth == 0) else "../" * depth
    targets = (
        ("a[href]", "href"),
        ("link[href]", "href"),
        ("script[src]", "src"),
    )
    for selector, attr in targets:
        for node in doc.select(selector):
            if "@root/" in node[attr]:
                node[attr] = node[attr].replace("@root/", prefix)


def do_toc_lists(doc, source_path):
    """Fix 'chapters' and 'appendices' lists."""
    for kind in ("chapters", "appendices"):
        selector = f"ol.{kind} ol"
        for node in doc.select(selector):
            node.parent.replace_with(node)
            node["class"] = node.get("class", []) + [kind]


def make_output_path(output_dir, renames, source_path):
    """Build output path."""
    if source_path.name in renames:
        source_path = Path(source_path.parent, renames[source_path.name])
    source_path = Path(str(source_path).replace(".md", ".html"))
    return Path(output_dir, source_path)


def parse_args(parser):
    """Parse command-line arguments."""
    parser.add_argument("--config", type=str, default="pyproject.toml", help="optional configuration file")
    parser.add_argument("--css", type=str, help="CSS file")
    parser.add_argument("--icon", type=str, help="icon file")
    parser.add_argument("--out", type=str, default="docs", help="output directory")
    parser.add_argument("--root", type=str, default=".", help="root directory")
    parser.add_argument("--templates", type=str, default="templates", help="templates directory")


def render_markdown(env, opt, source_path, content):
    """Convert Markdown to HTML."""
    template = choose_template(env, source_path)
    html = markdown(content, extensions=MARKDOWN_EXTENSIONS)
    html = template.render(content=html, css_file=opt.css, icon_file=opt.icon)

    transformers = (
        do_bibliography_links,
        do_glossary_links,
        do_markdown_links,
        do_tables,
        do_title,
        do_toc_lists,
        do_root_path_prefix, # must be last
    )
    doc = BeautifulSoup(html, "html.parser")
    for func in transformers:
        func(doc, source_path)

    return doc


if __name__ == "__main__":
    opt = parse_args(argparse.ArgumentParser()).parse_args()
    render(opt)



"""Slice sections from source files."""

import argparse
import ast
import esprima
from pathlib import Path


class MatchVisitor(esprima.NodeVisitor):
    """Visit Esprima nodes to find identifiers."""

    def __init__(self, expr):
        """Store expression to match."""
        super().__init__()
        self._expr = expr
        self._found = None

    def visit_VariableDeclarator(self, obj):
        """Variable declaration."""
        if obj.id.name == self._expr:
            self._found = obj

    def visit_FunctionDeclaration(self, obj):
        """Function declaration."""
        if obj.id.name == self._expr:
            self._found = obj


def main():
    """Main driver for testing."""
    options = parse_args()
    text = slice_file(options.src, options.expr)
    assert text is not None, f"cannot match {options.expr} in {options.src}"
    print(text)


def js_slice(src, expr):
    """Find and slice in JavaScript file."""
    text = Path(src).read_text()
    doc = esprima.parseScript(text, options={"loc": True})
    matcher = MatchVisitor(expr)
    matcher.visit(doc)
    node = matcher._found
    if node is None:
        return None
    return slice_lines(text, node.loc.start.line, node.loc.end.line)


def py_slice(src, expr):
    """Find and slice in Python file."""
    text = Path(src).read_text()
    doc = ast.parse(text)
    node = py_match(doc, expr)
    if node is None:
        return None
    return slice_lines(text, node.lineno, node.end_lineno)


def py_match(node, expr):
    """Find an entity by name."""
    if hasattr(node, "name") and (node.name == expr):
        return node
    if not hasattr(node, "body"):
        return None
    for child in node.body:
        found = py_match(child, expr)
        if found:
            return found
    return None


def parse_args():
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument("--expr", type=str, required=True, help="pattern expression")
    parser.add_argument("--src", type=str, required=True, help="source file")
    return parser.parse_args()


def slice_file(filepath, expr):
    """Extract slide from file."""
    language = Path(filepath).suffix.lstrip(".")
    handler = {
        "js": js_slice,
        "py": py_slice
    }.get(language, None)
    assert handler is not None, f"unknown file type {filepath}"
    return handler(filepath, expr)


def slice_lines(text, start, end):
    """Slice text from 1-based start to 1-based end line."""
    lines = text.split("\n")
    return "\n".join(lines[start - 1:end])


if __name__ == "__main__":
    main()



'''Interface for command-line script.'''

import argparse
import importlib.metadata
import importlib.resources
from pathlib import Path
import sys

from .lint import lint, parse_args as lint_parser
from .render import render, parse_args as render_parser
from .stats import stats, parse_args as stats_parser


INSTALL_FILES = (
    'templates/page.html',
    'templates/slides.html',
    'static/page.css',
    'static/slides.css',
    'static/slides.js',
)


def main():
    '''Main driver.'''
    parser = argparse.ArgumentParser()
    parser.add_argument('--version', action='store_true', help='show version')

    subparsers = parser.add_subparsers(dest='cmd')
    install_parser(subparsers.add_parser('install', help='install files'))
    lint_parser(subparsers.add_parser('lint', help='check site'))
    render_parser(subparsers.add_parser('render', help='build site'))
    stats_parser(subparsers.add_parser('stats', help='show stats'))

    opt = parser.parse_args()
    if opt.version:
        print(importlib.metadata.version('mccole'))
    elif opt.cmd == 'install':
        install(opt)
    elif opt.cmd == 'lint':
        lint(opt)
    elif opt.cmd == 'render':
        render(opt)
    elif opt.cmd == 'stats':
        stats(opt)
    else:
        print(f'unknown command {opt.cmd}', file=sys.stderr)
        sys.exit(1)


def install(opt):
    '''Install package files.'''
    outdir = Path(opt.root)
    outdir.mkdir(parents=True, exist_ok=True)
    root = importlib.resources.files(__name__.split('.')[0])
    mapping = {
        root.joinpath(filename): outdir.joinpath(Path(filename))
        for filename in INSTALL_FILES
    }

    exists = {str(dst) for dst in mapping.values() if dst.exists()}
    if exists and (not opt.force):
        print(f'not overwriting {", ".join(sorted(exists))} (use --force)', file=sys.stderr)
        sys.exit(1)

    for src, dst in mapping.items():
        dst.parent.mkdir(parents=True, exist_ok=True)
        dst.write_bytes(src.read_bytes())
    

def install_parser(parser):
    '''Parse command-line arguments.'''
    parser.add_argument('--force', action='store_true', help='overwrite')
    parser.add_argument('--root', type=str, default='.', help='root directory')
    return parser
